#include <assert.h> 
#include <stdio.h>
#include <string.h>

#include "_scm.h"
#include "pairs.h"
#include "gc.h"
#include "private-gc.h"



#define SCM_GC_CARD_BVEC_SIZE_IN_LONGS \
    ((SCM_GC_CARD_N_CELLS + SCM_C_BVEC_LONG_BITS - 1) / SCM_C_BVEC_LONG_BITS)
#define SCM_GC_IN_CARD_HEADERP(x) \
  (scm_t_cell *) (x) <  SCM_GC_CELL_CARD (x) + SCM_GC_CARD_N_HEADER_CELLS


size_t scm_max_segment_size;




scm_t_heap_seg_data *
scm_make_empty_heap_segment (scm_t_freelist *fl)
{
  scm_t_heap_seg_data * shs = malloc (sizeof (scm_t_heap_seg_data));

  if (!shs)
    {
      fprintf (stderr, "scm_get_new_heap_segment: out of memory.\n");
      abort ();
    }
  
  shs->bounds[0] = NULL;
  shs->bounds[1] = NULL;
  shs->malloced = NULL;
  shs->span = fl->span;
  shs->freelist  = fl;
  shs->free_cells = SCM_EOL;
  return shs;
}

void
scm_heap_segment (scm_t_heap_seg_data *h)
{
  free (h->malloced);
  free (h);
}


/*
  Fill SEGMENT with memory both for data and mark bits.

  RETURN:  1 on success, 0 failure  
 */
int 
scm_initialize_heap_segment_data (scm_t_heap_seg_data * segment, size_t requested)
{
  /*
    round upwards
   */
  int card_data_cell_count = (SCM_GC_CARD_N_CELLS - SCM_GC_CARD_N_HEADER_CELLS);
  int card_count =1 + (requested / sizeof (scm_t_cell)) /  card_data_cell_count; 

  /*
    one card extra due to alignment
  */
  size_t mem_needed = (1+card_count) * SCM_GC_SIZEOF_CARD
    + SCM_GC_CARD_BVEC_SIZE_IN_LONGS * card_count * SCM_SIZEOF_LONG
    ;
  scm_t_c_bvec_long * bvec_ptr = 0;
  scm_t_cell *  memory = 0;
  SCM_SYSCALL (memory = (scm_t_cell * ) calloc (1, mem_needed));

  if (memory == NULL)
    return 0;

  segment->malloced = memory;
  segment->bounds[0] = SCM_GC_CARD_UP (memory);
  segment->bounds[1] = segment->bounds[0] + card_count * SCM_GC_CARD_N_CELLS;

  segment->freelist->heap_size += scm_segment_cell_count (segment);
  
  bvec_ptr = (scm_t_c_bvec_long*) segment->bounds[1];
  {
    scm_t_cell *  ptr  = segment->bounds [0];
    for (;
	 ptr < segment->bounds[1]; ptr += SCM_GC_CARD_N_CELLS)
      {
	SCM_GC_CELL_BVEC (ptr) = bvec_ptr;
	if (segment->span == 2) 
	  SCM_GC_SET_CARD_DOUBLECELL (ptr);
 
	bvec_ptr += SCM_GC_CARD_BVEC_SIZE_IN_LONGS;
	scm_init_card_freelist (ptr, &segment->free_cells, segment->span);
      }
  }
  return 1;
}

int
scm_segment_card_count (scm_t_heap_seg_data * seg)
{
  return (seg->bounds[1] - seg->bounds[0]) / SCM_GC_CARD_N_CELLS;
}

/*
  Return the number of available single-cell data cells. 
 */
int
scm_segment_cell_count (scm_t_heap_seg_data * seg)
{
  return scm_segment_card_count (seg) * (SCM_GC_CARD_N_CELLS - SCM_GC_CARD_N_HEADER_CELLS)
    + ((seg->span == 2) ? -1 : 0);
}

void
scm_segment_clear_mark_space (scm_t_heap_seg_data *seg)
{
  scm_t_cell *  markspace = seg->bounds[1];

  memset (markspace, 0x00,
	  scm_segment_card_count (seg) *  SCM_GC_CARD_BVEC_SIZE_IN_LONGS * SCM_SIZEOF_LONG);
}

/*
  RETURN:

  number of cells collected.
*/
int
scm_sweep_segment (scm_t_heap_seg_data *seg)
{
  scm_t_cell *  ptr = seg->bounds[0];
  int free_cell_count = 0;


  /*
    Since only one master free_list is exhausted for every GC, the
    other one still contains cells. It would be a interesting
    optimization to reuse the tail of the free_cell list (eg. by
    marking free cells beforehand), then we can skip over the
    sweep phase.

    On the plus side this gets us a nicely ordered list of free cells
    again.
    
   */
  seg->free_cells = SCM_EOL; 
  for (; ptr < seg->bounds[1]; ptr += SCM_GC_CARD_N_CELLS)
    {
      free_cell_count += scm_sweep_card (ptr, &seg->free_cells, seg->span);
    }
#ifdef PARANOIA 
  validate_free_cells_list (seg->free_cells);
#endif
  return free_cell_count;
}




/*
  Heap segment table.

  The table is sorted by the address of the data itself. This makes for easy lookups

  perhaps it is worthwhile to remove the 2nd level of indirection in
  the table, but this certainly makes for cleaner code.
*/
scm_t_heap_seg_data ** scm_heap_segment_table;
size_t scm_heap_segment_table_size;
scm_t_cell *lowest_cell;
scm_t_cell *highest_cell; 



int
scm_next_segment_with_span (int idx,  int span)
{
  idx ++;
  
  while (idx <  scm_heap_segment_table_size)
    {
      if (scm_heap_segment_table[idx]->span == span)
	return idx;
      
      idx++;
    }

  return -1;
}


void
scm_clear_mark_space (void)
{
  int i = 0;
  for (; i < scm_heap_segment_table_size; i++)
    {
      scm_segment_clear_mark_space (scm_heap_segment_table[i]);
    }
}


/*
  RETURN: index of inserted segment.
 */
int
scm_insert_segment (scm_t_heap_seg_data * seg)
{
  size_t size = (scm_heap_segment_table_size + 1) * sizeof (scm_t_heap_seg_data *);
  SCM_SYSCALL(scm_heap_segment_table = ((scm_t_heap_seg_data **)
			       realloc ((char *)scm_heap_segment_table, size)));

  /*
    We can't alloc 4 more bytes. This is hopeless.
   */
  if (!scm_heap_segment_table)
    {
      fprintf (stderr, "scm_get_new_heap_segment: Could not grow heap segment table.\n");
      abort ();
    }




  if (!lowest_cell)
    {
      lowest_cell = seg->bounds[0];
      highest_cell = seg->bounds[1];
    }
  else
    {
      lowest_cell = SCM_MIN (lowest_cell, seg->bounds[0]);
      highest_cell = SCM_MAX (highest_cell, seg->bounds[1]);
    }

  
  {
    int i = 0;
    int j = 0;

    while (i < scm_heap_segment_table_size
	 && SCM_PTR_LE (scm_heap_segment_table[i]->bounds[0], seg->bounds[0]))
      i++;
    for (j = scm_heap_segment_table_size; j > i; --j)
      scm_heap_segment_table[j] = scm_heap_segment_table[j - 1];

    scm_heap_segment_table [i] = seg;
    scm_heap_segment_table_size ++;

    return i;
  }
}

void
scm_sweep_segments (void)
{
  int i = 0;
  for (; i < scm_heap_segment_table_size; i++)
    {
      scm_t_heap_seg_data * seg = scm_heap_segment_table[i];
      seg->freelist->collected += scm_sweep_segment (seg);
    }
}


/*
  Determine whether the given value does actually represent a cell in
  some heap segment.  If this is the case, the number of the heap
  segment is returned.  Otherwise, -1 is returned.  Binary search is
  used to determine the heap segment that contains the cell.


  I think this function is too long to be inlined. --hwn
*/
long int
scm_find_heap_segment_containing_object (SCM obj)
{
  if (!CELL_P (obj))
    return -1;
#if 1
  if ((scm_t_cell* ) obj < lowest_cell || (scm_t_cell*) obj >= highest_cell)
    return -1;
#endif  
  
    {
      scm_t_cell *  ptr = SCM2PTR (obj);
      unsigned long int i = 0;
      unsigned long int j = scm_heap_segment_table_size - 1;

      if (ptr < scm_heap_segment_table[i]->bounds[0])
	return -1;
      else if (scm_heap_segment_table[j]->bounds[1] <= ptr)
	return -1;
      else
	{
	  while (i < j)
	    {
	      if (ptr < scm_heap_segment_table[i]->bounds[1])
		{
		  break;
		}
	      else if (scm_heap_segment_table[j]->bounds[0] <= ptr)
		{
		  i = j;
		  break;
		}
	      else
		{
		  unsigned long int k = (i + j) / 2;

		  if (k == i)
		    return -1;
		  else if (ptr <  scm_heap_segment_table[k]->bounds[1])
		    {
		      j = k;
		      ++i;
		      if (ptr <  scm_heap_segment_table[i]->bounds[0])
			return -1;
		    }
		  else if (scm_heap_segment_table[k]->bounds[0] <= ptr)
		    {
		      i = k;
		      --j;
		      if (scm_heap_segment_table[j]->bounds[1] <= ptr)
			return -1;
		    }
		}
	    }

	  if (!DOUBLECELL_ALIGNED_P (obj) && scm_heap_segment_table[i]->span == 2)
	    return -1;
	  else if (SCM_GC_IN_CARD_HEADERP (ptr))
	    return -1;
	  else
	    return i;
	}
    }
}


/*
  Important entry point: try to grab some memory, and make it into a
  segment.

  RETURN: the index of the segment.
 */
int 
scm_get_new_heap_segment (scm_t_freelist *freelist, policy_on_error error_policy)
{
  size_t len;

  if (scm_gc_heap_lock)
    {
      /* Critical code sections (such as the garbage collector) aren't
       * supposed to add heap segments.
       */
      fprintf (stderr, "scm_get_new_heap_segment: Can not extend locked heap.\n");
      abort ();
    }


  /* Pick a size for the new heap segment.
   * The rule for picking the size of a segment is explained in
   * gc.h
   */
  {
    /* Assure that the new segment is predicted to be large enough.
     *
     * New yield should at least equal GC fraction of new heap size, i.e.
     *
     *   y + dh > f * (h + dh)
     *
     *    y : yield
     *    f : min yield fraction
     *    h : heap size
     *   dh : size of new heap segment
     *
     * This gives dh > (f * h - y) / (1 - f)
     */

    /*
      where is is this explanation supposed to be? 
     */
    int f = freelist->min_yield_fraction;
    unsigned long h = SCM_HEAP_SIZE;
    size_t min_cells = (f * h - 100 * (long) scm_gc_yield) / (99 - f);
    len =  SCM_EXPHEAP (freelist->heap_size);
#ifdef DEBUGINFO
    fprintf (stderr, "(%ld < %ld)", (long) len, (long) min_cells);
#endif

    /*
      Original code adds freelist->cluster_size here.
     */
    if (len < min_cells)
      len = min_cells;  
    len *= sizeof (scm_t_cell);
    /* force new sampling */
    freelist->collected = LONG_MAX;
  }

  if (len > scm_max_segment_size)
    len = scm_max_segment_size;

  {
    size_t smallest;
    scm_t_heap_seg_data * seg = scm_make_empty_heap_segment (freelist);
    
    smallest = 1024 * 10;  /* UGH.  */

    if (len < smallest)
      len = smallest;

    /* Allocate with decaying ambition. */
    while ((len >= SCM_MIN_HEAP_SEG_SIZE)
	   && (len >= smallest))
      {
	if (scm_initialize_heap_segment_data (seg, len))
	  {
	    return scm_insert_segment (seg);
	  }
	
	len /= 2;
      }
  }

  if (error_policy == abort_on_error)
    {
      fprintf (stderr, "scm_get_new_heap_segment: Could not grow heap.\n");
      abort ();
    }
  return -1;
}


void
scm_make_initial_segment (size_t init_heap_size, scm_t_freelist *freelist)
{
  scm_t_heap_seg_data * seg = scm_make_empty_heap_segment (freelist);
 
  if (scm_initialize_heap_segment_data (seg, init_heap_size))
    {
      scm_insert_segment (seg);
    }

  /*
    Why the fuck  try twice? --hwn
   */
  if (!seg->malloced)
    {
      scm_initialize_heap_segment_data (seg, SCM_HEAP_SEG_SIZE);
    }

  if (freelist->min_yield_fraction)
    freelist->min_yield = (freelist->heap_size * freelist->min_yield_fraction
			    / 100);
  
  {
    int grow_heap_p = (freelist->heap_size < freelist->min_yield);
    /*
      see comment in gc_sweep_freelist_finish 
     */

    if (grow_heap_p)
      scm_get_new_heap_segment (freelist, return_on_error);
  }
}


void
scm_gc_init_segments (void)
{
  size_t init_heap_size_1
    = scm_getenv_int ("GUILE_INIT_SEGMENT_SIZE_1", scm_default_init_heap_size_1);
  size_t init_heap_size_2
    = scm_getenv_int ("GUILE_INIT_SEGMENT_SIZE_2", scm_default_init_heap_size_2);
  
  scm_max_segment_size = scm_getenv_int ("GUILE_MAX_SEGMENT_SIZE", scm_default_max_segment_size);

  scm_make_initial_segment (init_heap_size_1, &scm_master_freelist);
  scm_make_initial_segment (init_heap_size_2, &scm_master_freelist2);
}

